# pylint: disable=missing-docstring,no-self-use,line-too-long

import unittest
from unittest.mock import patch, call, MagicMock
from typing import NamedTuple, Optional, Dict
from json import dumps
from urllib.parse import urlparse, parse_qs
import numpy  # type: ignore
import pandas  # type: ignore
from soil.data_structure import DS_NAMESPACE
import soil


class MockHttpResponse(NamedTuple):
    ''' Soil configuration class '''
    status_code: int
    text: str


# pylint: disable=unused-argument
def mock_http_post(url: str,
                   headers: Optional[Dict[str, str]] = None,
                   json: Optional[Dict[str, str]] = None
                   ) -> MockHttpResponse:
    assert json is not None
    response = {
        '_id': 'mock_id',
        'type': json['type']
    }
    return MockHttpResponse(status_code=201, text=dumps(response))


# pylint: disable=unused-argument, too-many-return-statements
def mock_http_get(url: str,
                  headers: Optional[Dict[str, str]] = None
                  ) -> MockHttpResponse:
    url_parts = urlparse(url)
    query_params = parse_qs(url_parts.query)
    if url_parts.path == '/v2/states/' and \
            query_params['name'][0] == 'non_existing_alias_but_valid_id':
        return MockHttpResponse(status_code=200, text=dumps([]))
    if url_parts.path == '/v2/states/' and \
            query_params['name'][0] == 'invalid_alias_invalidid':
        return MockHttpResponse(status_code=200, text=dumps([]))
    if url_parts.path == '/v2/states/' and \
            query_params['name'][0] == 'valid_alias':
        return MockHttpResponse(status_code=200, text=dumps([{'_id': 'alias_id'}]))
    if url_parts.path == '/v2/states/alias_id/':
        result = {'state': {'result_id': 'non_existing_alias_but_valid_id'}}
        return MockHttpResponse(status_code=200, text=dumps(result))
    if url_parts.path == '/v2/results/non_existing_alias_but_valid_id/':
        return MockHttpResponse(status_code=200, text=dumps({
            '_id': 'mock_id',
            'type': 'data_structures.some.test.type'
        }))
    if url_parts.path == '/v2/results/invalid_alias_invalidid/':
        return MockHttpResponse(status_code=404, text=dumps({
            'message': 'Document with id invalid_alias_invalidid not found'
        }))
    raise Exception('mock http case not found')


@patch('soil.api.requests.get', side_effect=mock_http_get)
@patch('soil.api.requests.post', side_effect=mock_http_post)
class TestData(unittest.TestCase):
    def test_upload_data_object_list(self, mock_post: MagicMock, mock_get: MagicMock) -> None:
        data = soil.data([1, 2, 3, 4])
        assert data.id == 'mock_id'
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + 'list.List'
        assert object.__getattribute__(data, 'data') is None
        assert object.__getattribute__(data, 'metadata') is None
        assert data.pipeline is None
        mock_post.assert_called_once_with('http://test_host.test/v2/results/', headers={'Authorization': 'Bearer test_token'}, json={'type': 'soil.data_structures.predefined.list.List', 'data': '[1, 2, 3, 4]', 'metadata': None})  # NOQA
        assert mock_get.call_count == 0

    def test_upload_data_object_dict(self, mock_post: MagicMock, mock_get: MagicMock) -> None:
        data = soil.data({'test': 'dictionary', 'value': 11})
        assert data.id == 'mock_id'
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + 'dict.Dict'
        assert object.__getattribute__(data, 'data') is None
        assert object.__getattribute__(data, 'metadata') is None
        assert data.pipeline is None
        mock_post.assert_called_once_with('http://test_host.test/v2/results/', headers={'Authorization': 'Bearer test_token'}, json={'type': 'soil.data_structures.predefined.dict.Dict', 'data': '{"test": "dictionary", "value": 11}', 'metadata': None})  # NOQA
        assert mock_get.call_count == 0

    def test_upload_data_object_numpy(self, mock_post: MagicMock, mock_get: MagicMock) -> None:
        data = soil.data(numpy.array([[1, 2], [3, 4]]))
        assert data.id == 'mock_id'
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + 'ndarray.Ndarray'
        assert object.__getattribute__(data, 'data') is None
        assert object.__getattribute__(data, 'metadata') is None
        assert data.pipeline is None
        mock_post.assert_called_once_with('http://test_host.test/v2/results/', headers={'Authorization': 'Bearer test_token'}, json={'type': 'soil.data_structures.predefined.ndarray.Ndarray', 'data': '[[1, 2], [3, 4]]', 'metadata': None})  # NOQA
        assert mock_get.call_count == 0

    def test_upload_data_object_pandas(self, mock_post: MagicMock, mock_get: MagicMock) -> None:
        data = soil.data(pandas.DataFrame(data={'col1': [1, 2], 'col2': [3, 4]}))
        assert data.id == 'mock_id'
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + 'pd_data_frame.PdDataFrame'
        assert object.__getattribute__(data, 'data') is None
        assert object.__getattribute__(data, 'metadata') is None
        assert data.pipeline is None
        mock_post.assert_called_once_with('http://test_host.test/v2/results/', headers={'Authorization': 'Bearer test_token'}, json={'type': 'soil.data_structures.predefined.pd_data_frame.PdDataFrame', 'data': '{"col1":{"0":1,"1":2},"col2":{"0":3,"1":4}}', 'metadata': None})  # NOQA
        assert mock_get.call_count == 0

    def test_get_data_valid_id(self, mock_post: MagicMock, mock_get: MagicMock) -> None:
        data = soil.data('non_existing_alias_but_valid_id')
        assert data.id == 'mock_id'
        assert data.sym_id is None
        assert data.dstype == 'data_structures.some.test.type'
        assert object.__getattribute__(data, 'data') is None
        assert object.__getattribute__(data, 'metadata') is None
        assert data.pipeline is None
        assert mock_post.call_count == 0
        assert mock_get.call_count == 2
        mock_get.assert_has_calls([
            call('http://test_host.test/v2/states/?name=non_existing_alias_but_valid_id',
                 headers={'Authorization': 'Bearer test_token'}),
            call('http://test_host.test/v2/results/non_existing_alias_but_valid_id/',
                 headers={'Authorization': 'Bearer test_token'})
        ])

    def test_get_data_valid_alias(self, mock_post: MagicMock, mock_get: MagicMock) -> None:
        data = soil.data('valid_alias')
        data = soil.data('non_existing_alias_but_valid_id')
        assert data.id == 'mock_id'
        assert data.sym_id is None
        assert data.dstype == 'data_structures.some.test.type'
        assert object.__getattribute__(data, 'data') is None
        assert object.__getattribute__(data, 'metadata') is None
        assert data.pipeline is None
        assert mock_post.call_count == 0
        assert mock_get.call_count == 5
        mock_get.assert_has_calls([
            call('http://test_host.test/v2/states/?name=valid_alias', headers={'Authorization': 'Bearer test_token'}),
            call('http://test_host.test/v2/states/alias_id/', headers={'Authorization': 'Bearer test_token'}),
            call('http://test_host.test/v2/results/non_existing_alias_but_valid_id/',
                 headers={'Authorization': 'Bearer test_token'}),
            call('http://test_host.test/v2/states/?name=non_existing_alias_but_valid_id',
                 headers={'Authorization': 'Bearer test_token'}),
            call('http://test_host.test/v2/results/non_existing_alias_but_valid_id/',
                 headers={'Authorization': 'Bearer test_token'})
        ])

    def test_get_data_invalid_alias(self, mock_post: MagicMock, mock_get: MagicMock) -> None:
        self.assertRaises(Exception, soil.data, 'invalid_alias_invalidid')
        assert mock_post.call_count == 0
        assert mock_get.call_count == 2
        mock_get.assert_has_calls([
            call('http://test_host.test/v2/states/?name=invalid_alias_invalidid',
                 headers={'Authorization': 'Bearer test_token'}),
            call('http://test_host.test/v2/results/invalid_alias_invalidid/',
                 headers={'Authorization': 'Bearer test_token'})
        ])



"""This module implements SOIL's Command LIne Interface"""
import argparse
import sys
import os
import json
import getpass
import zipfile
from json import JSONDecodeError    # Flake8...
import subprocess   # nosec
import virtualenv   # type: ignore
import requests

# pylint: disable=global-statement

env = dict()    # type: ignore
config = dict()

parser = argparse.ArgumentParser(prog='soil')
subparsers = parser.add_subparsers(dest='command')

configure_cmd = subparsers.add_parser('configure', help='Configure SOIL-CLI')
configure_cmd.add_argument('--reset', action='store_true', help="Reset the current configuration")

login_cmd = subparsers.add_parser('login', help='Login to configured soil instance')

init_cmd = subparsers.add_parser('init', help='Initialize a directory as a soil project')
init_cmd.add_argument('dirname', nargs='*', metavar='DIR', default='.',
                      help="Name of a directory to initialize as a soil app")

install_cmd = subparsers.add_parser('install', help='Installs the provided packages in the current soil project')
install_cmd.add_argument('packages', nargs='+', metavar='PACKAGE[S]', help="Package[s] to install")

run_cmd = subparsers.add_parser('run', help='Runs the provided module in the current soil project')
run_cmd.add_argument('module', nargs='+', metavar='module', help="Module to run")


def save_environment() -> None:
    """Saves the environment keys to a file, backing it up to prevent data loss."""

    global env  # pylint: disable=C0103
    os.rename(os.path.expanduser('~/.soil/soil.env'), os.path.expanduser('~/.soil/soil.env.bak'))
    env_file = open(os.path.expanduser('~/.soil/soil.env'), 'w')
    env_file.write(json.dumps(env, indent=4, sort_keys=True))
    env_file.close()
    os.remove(os.path.expanduser('~/.soil/soil.env.bak'))


def get_soil_root(relpath: str):    # type: ignore
    """Checks if the current dir is under a soil environment and returns its root. Returns None otherwise."""

    path = os.path.abspath(relpath) + '/'

    while path != "/":
        path, _ = os.path.split(path)
        if 'soil.yml' in os.listdir(path):
            return path

    return None


def soil_init() -> None:
    """Loads configuration and environment."""

    global env, config  # pylint: disable=C0103
    try:
        config_file = open(os.path.expanduser('~/.soil/soil.conf'), 'r')
        config = json.loads(config_file.read())
        config_file.close()
    except (IOError, JSONDecodeError):
        if sys.argv[1] != "configure":
            try:
                os.rename(os.path.expanduser('~/.soil/soil.conf.bak'), os.path.expanduser('~/.soil/soil.conf'))
            except IOError:
                print("Can not load soil configuration. Plase run soil configure to configure it.")
                sys.exit()
    try:
        env_file = open(os.path.expanduser('~/.soil/soil.env'), 'r')
        env = json.loads(env_file.read())
        env_file.close()
    except (IOError, JSONDecodeError):
        if sys.argv[1] != "configure":
            try:
                os.rename(os.path.expanduser('~/.soil/soil.conf.bak'), os.path.expanduser('~/.soil/soil.conf'))
            except IOError:
                print("Can not load soil environment. Plase run soil configure to initialize it.")
                sys.exit()


def configure(args: argparse.Namespace) -> None:
    """
    Allows the user to provide the configuration parameters interactively and stores them into a file.
    """

    global env, config  # pylint: disable=C0103
    try:
        os.makedirs(os.path.expanduser('~/.soil/'))
    except FileExistsError:
        # directory already exists
        pass

    try:
        if args.reset:
            try:
                os.rename(os.path.expanduser('~/.soil/soil.conf'), os.path.expanduser('~/.soil/soil.conf.bak'))
                os.rename(os.path.expanduser('~/.soil/soil.env'), os.path.expanduser('~/.soil/soil.env.bak'))
            except FileNotFoundError:
                print("Seems there is no previous configuration. Performing a clean config...")
                args.reset = False
        config_file = open(os.path.expanduser('~/.soil/soil.conf'), 'x')
        config["soil_url"] = input("Enter url of your soil instance: ")  # nosec - Input is safe in python3
        auth_url_msg = "Enter authentication provider URL: [https://auth.amalfianalytics.com] "
        config["auth_url"] = input(auth_url_msg)   # nosec - Input is safe in python3
        if config["auth_url"] == "":
            config["auth_url"] = "https://auth.amalfianalytics.com"
        config["auth_app_id"] = input("Enter your application id: ")  # nosec - Input is safe in python3
        config["auth_api_key"] = input("Enter your API key: ")  # nosec - Input is safe in python3
        config_file.write(json.dumps(config, indent=4, sort_keys=True))
        config_file.close()

        # Create the environment file
        env_file = open(os.path.expanduser('~/.soil/soil.env'), 'x')
        env_file.write("{\n}")
        env_file.close()

        if args.reset:
            os.remove(os.path.expanduser('~/.soil/soil.conf.bak'))
            os.remove(os.path.expanduser('~/.soil/soil.env.bak'))

    except FileExistsError:
        if not args.reset:
            print("Soil already configured (Use --reset to reconfigure):\n")
            for option, value in config.items():
                print(option + ": " + value)


def login() -> None:
    """
    Authenticates to the authentication backend and stores the credentials (JWT) in the environment
    """

    global env, config  # pylint: disable=C0103

    print(f"Authenticating to {config['auth_url']}...")
    username = input("Username: [" + env["auth"]["user"]["username"] + "]")  # nosec - Input is safe in python3
    if username == "":
        username = env["auth"]["user"]["username"]
    password = getpass.getpass()

    request_json = {'loginId': username, 'password': password, 'applicationId': config['auth_app_id']}

    # print(request_json)

    resp = requests.post(config['auth_url']+"/api/login",
                         headers={'Authorization': config['auth_api_key']},
                         json=request_json)

    if resp.status_code == 200 or resp.status_code == 202:
        env["auth"] = json.loads(resp.content)
        save_environment()
        print('Successfully logged in as ' + env['auth']['user']['username'] + "!")
    elif resp.status_code == 404:
        print("The user was not found or the password was incorrect.")
        login()


def init(args: argparse.Namespace) -> None:
    """
    Initializes the current directory or the one provided as argument as a SOIL project by placing
    all the directory structure and creating a python virtual environment.
    """

    dirname = vars(args)["dirname"][0]

    if dirname != '.':
        try:
            os.mkdir(dirname)
        except FileExistsError:
            print("The directory " + dirname + " already exists")
            sys.exit()

    if get_soil_root(dirname):
        if dirname != '.':
            os.rmdir(dirname)
        print("This location is already a soil project")
        sys.exit()
    else:
        with zipfile.ZipFile(os.path.dirname(os.path.realpath(__file__))+'/template.zip', 'r') as template_zip:
            template_zip.extractall(os.path.abspath('./'+dirname))
        # create and activate the virtual environment
        virtualenv.cli_run([os.path.abspath('./'+dirname+'/.venv')])


def install(args: argparse.Namespace) -> None:
    """
    Invokes current project's virtual environment pip to install the provided packages.
    Also updates requirements.txt
    """

    soil_root = get_soil_root('.')
    if not soil_root:
        print("This folder is not initalized as a soil project. Please run soil init to initialize it.")
        sys.exit()

    packages = vars(args)["packages"]

    for package in packages:
        subprocess.check_call(['.venv/bin/python', "-m", "pip", "install", package])    # nosec
    f = open('requirements.txt', 'w')
    f.write(os.popen(soil_root + '/.venv/bin/pip freeze').read())    # nosec
    f.close()


def run(args: argparse.Namespace) -> None:
    """
    Runs the script provided as argument using the python in the virtual environment
    """

    soil_root = get_soil_root('.')
    if not soil_root:
        print("This folder is not initalized as a soil project. Please run soil init to initialize it.")
        sys.exit()

    module = vars(args)["module"][0]

    os.chdir(soil_root)
    try:
        subprocess.run(['.venv/bin/python', '-m', 'soil_app.'+module], check=True)    # nosec
    except subprocess.CalledProcessError:
        sys.exit()


def main() -> None:
    """ main function"""

    args = parser.parse_args()

    soil_init()

    try:
        if args.command == 'configure':
            configure(args)
        elif args.command == 'login':
            login()
        elif args.command == 'init':
            init(args)
        elif args.command == 'install':
            install(args)
        elif args.command == 'run':
            run(args)
    except KeyboardInterrupt:
        print("\nKeyboardInterrupt - Exit...")
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0) # NOQA # pylint: disable=protected-access


if __name__ == "__main__":
    main()



from soil import modulify
from soil.data_structures.predefined.list import List


@modulify(output_types=lambda *inputs, **args: [List])
def even(input):
    return [List([i for i in input.data if i % 2 == 0], {})]



'''
Soil package.
'''
from soil.modulify import modulify
from soil.data import data
from soil.alias import alias
from soil import finder
from soil.logger import logger
import soil.connectors as connectors

finder.install()

__all__ = [
    'modulify',
    'data_structures',
    'modules',
    'data',
    'alias',
    'logger',
    'connectors'
]



from soil import modulify, logger
# from soil.data_structures.simple_datastructure import SimpleDataStructure
from soil.data_structures.es_datastructure import EsDataStructure


@modulify(output_types=lambda *inputs, **args: [EsDataStructure])
def simple_mean(patients, aggregation_column=None):
    logger.info('This is a log from simple_mean')
    logger.debug('This is another log from simple_mean')
    if aggregation_column is None:
        raise TypeError('Expected aggregation_column parameter')
    total_sum = 0
    count = 0
    for patient in patients:
        if hasattr(patient, aggregation_column):
            val = patient[aggregation_column]
            total_sum += val
            count += 1
    return [EsDataStructure({'mean': round(total_sum / count)}, metadata={'awww': 'yeah'})]



from soil.data_structures.data_structure import DataStructure
from soil.connectors.elastic_search import create_db_object


class EsDataStructure(DataStructure):
    @staticmethod
    def unserialize(_nothing, metadata, db_object):
        result = db_object.query({})
        data = next(result[0])
        MyDS = EsDataStructure(data, metadata, db_object)
        return MyDS

    def serialize(self):
        es_db_object = create_db_object(index='my-es-index')
        es_db_object.insert(self.data, id='first_id')
        return es_db_object

    def get_data(self, **args):
        return self.data



import json
from soil.data_structures.data_structure import DataStructure


class SimpleDataStructure(DataStructure):
    @staticmethod
    def unserialize(json_data, metadata):
        MyDS = SimpleDataStructure(json.loads(next(json_data)), metadata)
        return MyDS

    def serialize(self):
        return json.dumps(self.data)

    def get_data(self, **args):
        return self.data



# pylint: disable=missing-docstring,no-self-use,line-too-long
from typing import NamedTuple, Optional, Dict
import unittest
from unittest.mock import patch, call, MagicMock
from urllib.parse import urlparse, parse_qs
from json import dumps
import soil
from soil.data_structure import DataStructure


class MockHttpResponse(NamedTuple):
    ''' Soil configuration class '''
    status_code: int
    text: str


def set_id(self: DataStructure) -> None:
    self.id = 'mock_id'


# pylint: disable=unused-argument
def mock_http_patch(url: str,
                    headers: Optional[Dict[str, str]] = None,
                    json: Optional[Dict[str, str]] = None
                    ) -> MockHttpResponse:
    assert json is not None
    return MockHttpResponse(status_code=200, text=dumps({}))


# pylint: disable=unused-argument
def mock_http_post(url: str,
                   headers: Optional[Dict[str, str]] = None,
                   json: Optional[Dict[str, str]] = None
                   ) -> MockHttpResponse:
    assert json is not None
    url_parts = urlparse(url)
    if url_parts.path == '/v2/results/':
        response = {
            '_id': 'mock_id',
            'type': json['type']
        }
        return MockHttpResponse(status_code=201, text=dumps(response))
    if url_parts.path == '/v2/states/':
        return MockHttpResponse(status_code=200, text=dumps({}))
    raise Exception('mock http case not found')


# pylint: disable=unused-argument
def mock_http_get(url: str,
                  headers: Optional[Dict[str, str]] = None
                  ) -> MockHttpResponse:
    url_parts = urlparse(url)
    query_params = parse_qs(url_parts.query)
    if url_parts.path == '/v2/states/' and \
            query_params['name'][0] == 'test_alias':
        return MockHttpResponse(status_code=200, text=dumps([]))
    raise Exception('mock http case not found')


# pylint: disable=unused-argument
def mock_http_get2(url: str,
                   headers: Optional[Dict[str, str]] = None
                   ) -> MockHttpResponse:
    url_parts = urlparse(url)
    query_params = parse_qs(url_parts.query)
    if url_parts.path == '/v2/states/' and \
            query_params['name'][0] == 'valid_alias':
        return MockHttpResponse(status_code=200, text=dumps([{'_id': 'alias_id'}]))
    if url_parts.path == '/v2/states/alias_id/':
        result = {'_id': 'alias_id', 'state': {'result_id': 'non_existing_alias_but_valid_id'}}
        return MockHttpResponse(status_code=200, text=dumps(result))
    raise Exception('mock http case not found')


class TestAlias(unittest.TestCase):
    @patch('soil.api.requests.get', side_effect=mock_http_get)
    @patch('soil.api.requests.post', side_effect=mock_http_post)
    @patch('soil.api.requests.patch', side_effect=mock_http_patch)
    def test_alias_new(self, mock_patch: MagicMock, mock_post: MagicMock, mock_get: MagicMock) -> None:
        data = soil.data([1, 2, 3, 4])
        soil.alias('test_alias', data)
        mock_get.assert_has_calls([call('http://test_host.test/v2/states/?name=test_alias',
                                        headers={'Authorization': 'Bearer test_token'})])
        mock_post.assert_has_calls([
            call('http://test_host.test/v2/results/', headers={'Authorization': 'Bearer test_token'}, json={
                'type': 'soil.data_structures.predefined.list.List', 'data': '[1, 2, 3, 4]', 'metadata': None}),
            call('http://test_host.test/v2/states/', headers={'Authorization': 'Bearer test_token'}, json={
                'name': 'test_alias', 'state': {'alias': 'test_alias', 'result_id': 'mock_id'}})
        ])
        assert mock_patch.call_count == 0

    @patch('soil.api.requests.get', side_effect=mock_http_get2)
    @patch('soil.api.requests.post', side_effect=mock_http_post)
    @patch('soil.api.requests.patch', side_effect=mock_http_patch)
    def test_alias_existing(self, mock_patch: MagicMock, mock_post: MagicMock, mock_get: MagicMock) -> None:
        data = soil.data([1, 2, 3, 4])
        soil.alias('valid_alias', data)
        mock_get.assert_has_calls([
            call('http://test_host.test/v2/states/?name=valid_alias', headers={'Authorization': 'Bearer test_token'}),
            call('http://test_host.test/v2/states/alias_id/', headers={'Authorization': 'Bearer test_token'})
        ])
        mock_post.assert_has_calls([
            call('http://test_host.test/v2/results/', headers={'Authorization': 'Bearer test_token'}, json={
                'type': 'soil.data_structures.predefined.list.List', 'data': '[1, 2, 3, 4]', 'metadata': None})
        ])
        mock_patch.assert_has_calls([
            call('http://test_host.test/v2/states/alias_id/', headers={'Authorization': 'Bearer test_token'}, json={
                'name': 'valid_alias', 'state': {'alias': 'valid_alias', 'result_id': 'mock_id'}})
        ])

    @patch('soil.api.requests.get', side_effect=mock_http_get2)
    @patch('soil.api.requests.post', side_effect=mock_http_post)
    @patch('soil.api.requests.patch', side_effect=mock_http_patch)
    def test_alias_data_structure(self, mock_patch: MagicMock, mock_post: MagicMock, mock_get: MagicMock) -> None:
        with patch.object(DataStructure, 'get_data', side_effect=set_id, autospec=True) as mock_get_data:
            data = DataStructure(dstype='data_structures.some.test.type')
            soil.alias('valid_alias', data)
            assert mock_get_data.call_count == 1
        mock_get.assert_has_calls([
            call('http://test_host.test/v2/states/?name=valid_alias',
                 headers={'Authorization': 'Bearer test_token'}),
            call('http://test_host.test/v2/states/alias_id/', headers={'Authorization': 'Bearer test_token'})
        ])
        assert mock_post.call_count == 0
        mock_patch.assert_has_calls([
            call('http://test_host.test/v2/states/alias_id/', headers={'Authorization': 'Bearer test_token'}, json={
                'name': 'valid_alias', 'state': {'alias': 'valid_alias', 'result_id': 'mock_id'}})
        ])



import soil
import logging
from soil.modules.even import even

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

data_ref = soil.data([i for i in range(50)])

soil.alias('my_data2', data_ref)

data_ref2 = soil.data('my_data2')
even_numbers, = even(data_ref2)
for i in even_numbers.data:
    print(i)



'''
This module defines the @modulify decorator.
'''
import random
from typing import List, Dict, Any, Callable, Optional, cast
from soil.data_structure import DataStructure
from soil.pipeline import Pipeline


def _generate_data_structure_ids(fn_name: str, quantity: int) -> List[str]:
    rnd = random.randint(1000, 9999)  # nosec
    return [fn_name.replace('.', '/') + '-' + str(rnd) + '-' + str(i) for i in range(quantity)]


def _generate_transformation(
        input_ids: List[str],
        output_ids: List[str],
        fn_name: str,
        args: Dict[str, Any]
) -> Dict[str, Any]:
    tid = fn_name + '-' + str(random.randint(1000, 9999))  # nosec
    return {
        'id': tid,
        'module': fn_name,
        'inputs': input_ids,
        'outputs': output_ids,
        'args': args
    }


def modulify(
        _func: Optional[Callable[..., List[DataStructure]]] = None,
        *,
        output_types: List[Callable] = None,
        num_outputs: int = 1,
        _from_db: bool = False
) -> Callable:
    '''Decorates a function to mark it as a soil module.'''
    def decorator(fn: Callable[..., List[DataStructure]]) -> Callable:
        fnn = fn if _func is None else _func
        mod_name = fnn.__module__ + '.' + fnn.__name__
        if not _from_db and output_types is None:
            raise ValueError('@modulify: ' + mod_name + ' missing output_types')

        def decorated(*inputs: DataStructure, **kwargs: Any) -> List[DataStructure]:
            # print('run', _func.__module__, _func.__name__)
            new_pipeline = Pipeline.merge_pipelines(*[input.pipeline for input in inputs if input.pipeline is not None])
            output_ids = _generate_data_structure_ids(mod_name, num_outputs)
            input_ids = [d.id if d.id is not None else d.sym_id for d in inputs]
            for i in input_ids:
                assert isinstance(i, str)
            transformation = _generate_transformation(
                input_ids=cast(List[str], input_ids),
                output_ids=output_ids,
                fn_name=mod_name,
                args=kwargs
            )
            new_pipeline = new_pipeline.add_transformation(transformation)
            outputs = [
                DataStructure(sym_id=sym_id, pipeline=new_pipeline)
                for sym_id in output_ids
            ]
            return outputs
        return decorated
    if _func is None:
        return decorator
    return decorator(_func)



'''
This package contains calls to the SOIL's REST API
'''
import logging
from typing import Dict, Any, Optional
import json
from urllib.parse import urlencode
import requests
from soil.configuration import GLOBAL_CONFIG
from soil.types import Plan, GetModule, Experiment, Result

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

API_ROOT = str(GLOBAL_CONFIG.host) + '/v2/'
HEADERS = {
    'Authorization': 'Bearer ' + str(GLOBAL_CONFIG.token)
}


def upload_data(dtype: str, data: Any, metadata: Any) -> Result:
    ''' Upload data to the cloud as a new dataset.'''
    logger.debug('upload_data:%s')
    url = API_ROOT + 'results/'
    body = {
        'type': dtype,
        'data': data,
        'metadata': metadata
    }
    response = requests.post(url, headers=HEADERS, json=body)
    if response.status_code != 201:
        raise ValueError(response.text)
    return json.loads(response.text)


def get_result(result_id: str) -> Dict[str, Any]:
    ''' Get the result data '''
    logger.debug('get_result:%s', result_id)
    url = API_ROOT + 'results/' + result_id + '/'
    response = requests.get(url, headers=HEADERS)
    if response.status_code != 200:
        raise Exception(response.text)
    return json.loads(response.text)


def get_result_data(result_id: str, query: Optional[Dict[str, str]] = None) -> Dict[str, Any]:
    ''' Get the result data '''
    if query is None:
        query = {}
    logger.debug('get_result:%s %s', result_id, str(query))
    url = API_ROOT + 'results/' + result_id + '/data/?' + urlencode(query)
    response = requests.get(url, headers=HEADERS)
    if response.status_code != 200:
        raise Exception(response.text)
    return json.loads(response.text)


def upload_module(module_name: str, code: str, is_package: bool) -> None:
    ''' Uploads a module '''
    logger.debug('upload_module:%s', module_name)
    url = API_ROOT + 'modules/'
    data = {
        'name': module_name,
        'code': code,
        'is_package': is_package
    }
    response = requests.post(url, headers=HEADERS, json=data)
    if response.status_code != 200:
        raise Exception(response.text)


def get_module(full_name: str) -> GetModule:
    ''' Downloads a module '''
    logger.debug('get_module:%s', full_name)
    url = API_ROOT + 'modules/' + full_name + '/'
    response = requests.get(url, headers=HEADERS)
    if response.status_code != 200:
        raise Exception(response.text)
    return json.loads(response.text)


def set_alias(alias: str, result_id: str) -> None:
    ''' Sets an alias for a result. Updates a previous one with the same name. '''
    logger.debug('set_alias: %s = %s', alias, result_id)
    obj = {
        'name': alias,
        'state': {
            'alias': alias,
            'result_id': result_id
        }
    }
    try:
        old_alias_id = get_alias(alias)['_id']
        url = API_ROOT + 'states/' + old_alias_id + '/'
        response = requests.patch(url, headers=HEADERS, json=obj)
    except KeyError:
        url = API_ROOT + 'states/'
        response = requests.post(url, headers=HEADERS, json=obj)
    if response.status_code != 200:
        raise ValueError('Failed to create alias ' + alias)


def get_alias(alias: str) -> Dict[str, Any]:
    ''' Gets an alias '''
    logger.debug('get_alias: %s', alias)
    url = API_ROOT + 'states/?name=' + alias
    response = requests.get(url, headers=HEADERS)
    if response.status_code != 200:
        raise ValueError('Error getting alias:' + response.text)
    aliases = json.loads(response.text)
    if len(aliases) == 0:
        raise KeyError('Alias not found')
    alias_id = aliases[0]['_id']
    url = API_ROOT + 'states/' + alias_id + '/'
    response = requests.get(url, headers=HEADERS)
    if response.status_code != 200:
        raise ValueError('Error getting alias with id ' + response.text)
    return json.loads(response.text)


def create_experiment(plan: Plan) -> Experiment:
    ''' Runs an experiment in SOIL '''
    logger.debug('create_experiment: %s', str(plan))
    url = API_ROOT + 'experiments/'
    experiment = {
        'name': '',
        'description': '',
        'plan': plan
    }
    response = requests.post(url, headers=HEADERS, json={'experiment': experiment})
    if response.status_code != 200:
        raise ValueError('Error creating the experiment:' + response.text)
    return json.loads(response.text)['experiment']


def get_experiment(experiment_id: str) -> Experiment:
    ''' Runs an experiment from SOIL '''
    logger.debug('get_experiment: %s', experiment_id)
    url = API_ROOT + 'experiments/' + experiment_id + '/'
    response = requests.get(url, headers=HEADERS)
    if response.status_code != 200:
        raise ValueError('Error getting the experiment ' + experiment_id + '\n' + response.text)
    return json.loads(response.text)


def get_experiment_logs(experiment_id: str, start_date: str) -> Any:
    ''' Gets logs from a SOIL experiment '''
    logger.debug('get_experiment_id: %s since %s', experiment_id, start_date)
    url = API_ROOT + 'experiments/' + experiment_id + '/logs/?' + urlencode({'start_date': start_date})
    response = requests.get(url, headers=HEADERS)
    if response.status_code != 200:
        raise ValueError('Error getting the experiment ' + experiment_id + '\n' + response.text)
    return json.loads(response.text)



''' Defines the configuration for Soil '''
from typing import Optional, NamedTuple
import logging
from os import getenv
import json
import requests
# TODO add windows support

# pylint: disable=invalid-name
env = getenv('PY_ENV', 'development')


def _refresh_token(auth_host: str, refresh_token: str) -> str:
    url = auth_host + '/api/jwt/refresh'
    response = requests.post(url, json={'refreshToken': refresh_token})
    if response.status_code != 200:
        raise ValueError('Invalid refresh token please run soil login again.')
    return json.loads(response.text)['token']


SOIL_URL = ''
if env != 'test':
    try:
        with open(getenv('HOME', '') + '/.soil/soil.conf') as conf_file:
            CONF = json.loads(conf_file.read())
            SOIL_URL = CONF['soil_url']
    except FileNotFoundError:
        logging.warning('~/.soil/soil.conf file not found. Please run soil configure.')
else:
    SOIL_URL = 'http://test_host.test'

TOKEN = ''  # nosec

if env != 'test':
    try:
        with open(getenv('HOME', '') + '/.soil/soil.env') as env_file:
            ENV = json.loads(env_file.read())
            TOKEN = _refresh_token(CONF['auth_url'], ENV['auth']['refreshToken'])
    except FileNotFoundError:
        logging.warning('~/.soil/soil.env file not found. Please run soil login.')
else:
    TOKEN = 'test_token'  # nosec

DEFAULT_CONFIG = {
    'host': getenv('SOIL_HOST', SOIL_URL),
    'token': getenv('SOIL_TOKEN', TOKEN)
}


class SoilConfiguration(NamedTuple):
    ''' Soil configuration class '''
    host: Optional[str]
    token: Optional[str]


GLOBAL_CONFIG = SoilConfiguration(**DEFAULT_CONFIG)


# Not used for now
# def config(token: Optional[str] = None, host: Optional[str] = None) -> SoilConfiguration:
#     ''' Set the Soil's configuration '''
#     global GLOBAL_CONFIG  # pylint: disable=global-statement
#     new_config = SoilConfiguration(token=token, host=host)
#     GLOBAL_CONFIG = SoilConfiguration(**{**GLOBAL_CONFIG._asdict(), **new_config._asdict()})
#     return GLOBAL_CONFIG



import logging
import soil # NOQA
from soil.modules.preprocessing.filters import row_filter # NOQA
from soil.modules.simple_module import simple_mean # NOQA
from soil.modules.statistics.statistics import Statistics # NOQA

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

patients = soil.data('5ece3f310af3725d0a7d76b6')
older, = row_filter.RowFilter(patients, age={'gte': 60})
women, = row_filter.RowFilter(older, sex={'eql': '1'})
# statistics, = Statistics(women, operations=[{
#             'operation_name': 'mean_age',
#             'fn': 'mean',
#             'aggregation_variable': {'name': 'age'},
#             'partition_variables': [{'name': 'sex'}]
#         }])
statistics, = simple_mean(women, aggregation_column='age')
print(statistics.data)
